////////////////////////////////
// GameMode_Agar.io/server.cs //
////////////////////////////////
// TABLE OF CONTENTS
// #1.0 global setting variables
// #2.0 datablocks
//   #2.1 (playerData) agarioPlayer
// #3.0 package Agario
//   #3.1 GameConnection::getSpawnPoint
//   #3.2 GameConnection::applyBodyColors
//   #3.3 Observer::onTrigger
// #4.0 main functionality
//   #4.1 agarioPlayer::onAdd
//   #4.2 agarioPlayer::onTrigger
//   #4.3 player::absorb
//   #4.4 player::agarioRescale
//   #4.5 player::absorbSched
//   #4.6 colorSaturate
//   #4.7 agarSpawnFood
// #5.0 bottomprint hud
// #6.0 setup
//   #6.1 spawn food

// #1.0
$Agario::StartingMass = mPow(0.5, 3);
$Agario::FoodMass = mPow(0.3, 3);
$Agario::AbsorbSchedLen = 100;
$Agario::EjectMass = mPow(0.5, 3);
$Agario::SplitTimeout = 15000;
$Agario::SplitRuntime = 2500;
$Agario::LeaderDisplayInterval = 60000;

// #2.0

// #2.1
datablock playerData(agarioPlayer : playerStandardArmor)
{
	canJet = false;
	airControl = 0;
	jumpForce = 0;
	runForce = 4320 * 2;
	maxStepHeight = 0;
	maxTools = 0;
	uiName = "Agario Player";
};

// #3.0
package Agario
{
	// #3.1
	function GameConnection::getSpawnPoint(%this)
	{
		return (getRandom(-100, 100)) SPC (getRandom(-100, 100)) SPC 50;
	}

	// #3.3
	function GameConnection::spawnPlayer(%this)
	{
		parent::spawnPlayer(%this);
		%this.player.setDatablock(agarioPlayer);	
	}

	// #3.2
	function GameConnection::applyBodyColors(%this)
	{
		if(isObject(%obj = %this.player))
		{
			%obj.setShapeNameColor(%this.chestColor);
		}
		parent::applyBodyColors(%this);	
	}

	// #3.3
	function serverCmdSuicide(%client)
	{
		%client.instantRespawn();
	}
};
activatePackage(Agario);

// #4.0

// #4.1
function agarioPlayer::onNewDatablock(%this, %obj)
{
	if(isObject(%client = %obj.client))
	{
		if(!isObject(%agarGroup = %client.agarGroup))
		{
			%agarGroup = new simGroup();
			missionCleanup.add(%agarGroup);
			%client.agarGroup = %agarGroup;
		}
		if(!isEventPending(%client.agarHudSched))
		{
			%client.agarHud();
			%client.ticker = "Jet to eject some mass. Jump to split. | If your camera messes up, ctrl+k to respawn.";
		}
		%agarGroup.schedule(1, add, %obj);
	}
	%obj.mass = $Agario::StartingMass;
	%obj.agarRescale();
	%obj.schedule($Agario::AbsorbSchedLen, absorbSched);
	parent::onNewDatablock(%this, %obj);
}

// #4.2
function agarioPlayer::onTrigger(%this, %obj, %slot, %val)
{
	if(%slot == 2 && %val) //split
	{
		if(%obj.mass >= $Agario::StartingMass * 2)
		{
			%obj.mass/= 2;
			%scale = getWord(%obj.getScale(), 2);
			%obj.agarRescale();
			%vec = vectorScale(%obj.getEyeVector(), 0.5 + %scale * 2);
			%pos = vectorAdd(vectorAdd(%obj.getPosition(), %vec), %obj.getVelocity());
			if(getWord(%pos, 2) < 0)
			{
				%pos = setWord(%pos, 2, 0);
			}
			%rot = rotFromTransform(%obj.getTransform());
			%split = new aiPlayer()
			{
				position = %pos;
				datablock = agarioPlayer;
				splitTimeout = getSimTime() + $Agario::SplitTimeout;
			};
			%split.mass = %obj.mass;
			if(isObject(%client = %obj.client))
			{
				%split.client = %client;
				%client.agarGroup.add(%split);
				%client.player = %split;
				%client.applyBodyParts();
				%client.applyBodyColors();
				%client.player = %obj;
			}
			%split.agarRescale();
			%split.setTransform(%pos SPC %rot);
			%split.setVelocity(vectorScale(%vec, 3));
			%split.setMoveY(1);
			%split.schedule($Agario::SplitRuntime, clearMoveY);
			%split.schedule($Agario::SplitRuntime, setMoveObject, %obj);
		}
	}
	else if(%slot == 4 && %val) //eject
	{
		if(%obj.mass >= $Agario::StartingMass + $Agario::EjectMass)
		{
			%obj.mass-= $Agario::EjectMass;
			%scale = getWord(%obj.getScale(), 2);
			%obj.agarRescale();
			%vec = vectorScale(%obj.getEyeVector(), 0.5 + %scale * 2);
			%pos = vectorAdd(vectorAdd(%obj.getPosition(), %vec), %obj.getVelocity());
			if(getWord(%pos, 2) < 0)
			{
				%pos = setWord(%pos, 2, 0);
			}
			%rot = rotFromTransform(%obj.getTransform());
			%ejection = new aiPlayer()
			{
				position = %pos;
				datablock = playerStandardArmor;
				splitTimeout = getSimTime() + 2 * $Agario::AbsorbSchedLen;
				ejection = true;
			};
			%ejection.mass = $Agario::EjectMass;
			if(isObject(%client = %obj.client))
			{
				%ejection.client = %client;
				%client.agarGroup.add(%ejection);
				%client.agarGroup.schedule(2 * $Agario::AbsorbSchedLen, remove, %ejection);
				%client.player = %ejection;
				%client.applyBodyParts();
				%client.applyBodyColors();
				%client.player = %obj;
			}
			%ejection.agarRescale();
			%ejection.setTransform(%pos SPC %rot);
			%ejection.setVelocity(vectorScale(%vec, 3));
		}
	}
	parent::onTrigger(%this, %obj, %slot, %val);
}

// #4.3
// %big absorbs %smol, deleting %smol (and handling controlobject transfer) and growing %big
function Player::absorb(%big, %smol)
{
	if(%smol.absorbed || !isObject(%smol) || !isObject(%big))
	{
		return;
	}
	if(%smol.food)
	{
		schedule(100, 0, agarSpawnFood);
	}
	%smol.absorbed = true;
	if(%smol.mass > 0)
	{
		%smolmass = %smol.mass;
	}
	else
	{
		%smolmass = mPow(getWord(%smol.getScale(), 2), 3);
	}
	%big.mass+= %smolmass;
	if(isObject(%client = %big.getControllingClient()))
	{
		for(%i = 0; %i < %client.agarGroup.getCount(); %i++)
		{
			%groupmass+= %client.agarGroup.getObject(%i).mass;
		}
		%client.setScore(mFloatLength(%groupmass, 0));
	}
	if(isObject(%client = %smol.getControllingClient()))
	{
		%agarGroup = %client.agarGroup;
		%biggestMass = 0;
		%biggest = 0;
		%agarGroup.remove(%smol);
		for(%i = 0; %i < %agarGroup.getCount(); %i++)
		{
			%obj = %agarGroup.getObject(%i);
			if(%obj.mass > %biggestMass)
			{
				%biggestMass = %obj.mass;
				%biggest = %obj;
			}
		}
		if(isObject(%biggest))
		{
			%client.setControlObject(%biggest);
			%client.player = %biggest;
			for(%i = 0; %i < %agarGroup.getCount(); %i++)
			{
				%obj = %agarGroup.getObject(%i);
				if(%obj != %biggest)
				{
					%obj.setMoveObject(%biggest);
				}
			}
		}
		else
		{
			%client.schedule(1, instantRespawn);
		}
	}
	%smol.spawnExplosion(spawnProjectile, getWord(%smol.getScale(), 2));
	%smol.delete();
	%big.agarRescale();
}

// #4.4
function Player::agarRescale(%obj)
{
	%scale = mFloatLength(mPow(%obj.mass, 0.333), 2);
	%obj.setScale(%scale SPC %scale SPC %scale);
	%speed = 10 / mPow(%scale, 1.5);
	%obj.setMaxForwardSpeed(%speed);
	%obj.setMaxSideSpeed(%speed);
	%obj.setMaxBackwardSpeed(%speed);
	%obj.setMaxCrouchForwardSpeed(%speed);
	%obj.setMaxCrouchSideSpeed(%speed);
	%obj.setMaxCrouchBackwardSpeed(%speed);
}

// #4.5
function Player::absorbSched(%obj)
{
	if(!isObject(%obj))
	{
		return;
	}
	%pos = %obj.getPosition();
	%scale = getWord(%obj.getScale(), 2);
	InitContainerRadiusSearch(%pos, 0.5 + %scale, $Typemasks::PlayerObjectType);
	while(%hit = ContainerSearchNext())
	{
		%dist = vectorDist(%pos, %hit.getPosition());
		if(%dist > %scale + 0.5)
		{
			continue;
		}
		if(%hit == %obj)
		{
			continue;
		}
		if(%hit.getGroup() == %obj.getGroup())
		{
			//same agargroup
			%time = getSimTime();
			if(%hit.splitTimeout >= %time || %obj.splitTimeout >= %time)
			{
				continue;
			}
			else
			{
				if(isObject(%obj.getControllingClient()))
				{
					%obj.absorb(%hit);
				}
				else if(isObject(%hit.getControllingClient()))
				{
					%hit.absorb(%obj);
				}
			}
		}
		%hscale = getWord(%hit.getScale(), 2);
		if(%scale * 0.9 >= %hscale)
		{
			%obj.absorb(%hit);
		}
		else if(%hscale * 0.9 >= %scale)
		{
			%hit.absorb(%obj);
		}
	}
	if(%obj.mass > 40)
	{
		%decay = 0.1 * mPow((%obj.mass - 40) / 20, 2);
		%decay/= 1000 / $Agario::AbsorbSchedLen;
		%obj.mass-= %decay;
		%obj.agarRescale();
	}
	%obj.schedule($Agario::AbsorbSchedLen, absorbSched, %obj);
}

// #4.6
function colorSaturate(%col)
{
	%r = getWord(%col, 0);
	%g = getWord(%col, 1);
	%b = getWord(%col, 2);
	if(%r > %g)
	{
		if(%r > %b)
		{
			%scale = %r;
		}
		else //%b > %r
		{
			%scale = %b;
		}
	}
	else //%g > %r
	{
		if(%g > %b)
		{
			%scale = %g;
		}
		else //%b > %g
		{
			%scale = %b;
		}
	}
	return vectorScale(%col, 1 / %scale);
}

// #4.7
function agarSpawnFood()
{
	%bot = new AiPlayer()
	{
		datablock = playerStandardArmor;
		position = getRandom(-100, 100) SPC getRandom(-100, 100) SPC 50;
		mass = $Agario::FoodMass;
		food = true;
	};
	if(clientGroup.getCount() > 0)
	{
		%client = clientGroup.getObject(getRandom(0, clientGroup.getCount() - 1));
		%obj = %client.player;
		%client.player = %bot;
		%client.applyBodyParts();
		%client.player = %obj;
	}
	%bot.agarRescale();
	%vec = vectorSub(getRandom() SPC getRandom() SPC 0, "-0.5 -0.5 0");
	%bot.setAimVector(%vec);
	%color = colorSaturate(getRandom() SPC getRandom() SPC getRandom());
	%bot.setNodeColor("ALL", %color SPC 1);
}

// #5.0
$tickerLength = 41;
$tickerSpeed = 250;
$tickerFormatting = "<font:Courier New Bold:28><color:FFFFFF>";

function GameConnection::agarHud(%client)
{
	if(isEventPending(%client.agarHudSched))
	{
		cancel(%client.agarHudSched);
	}
	for(%i = 0; %i < %client.agarGroup.getCount(); %i++)
	{
		%mass+= %client.agarGroup.getObject(%i).mass;
	}
	%mass = mFloatLength(%mass, 2);
	while(strLen(%mass) < 6)
	{
		%mass = " " @ %mass;
	}
	%mass = %mass @ " | ";
	if(%client.tickerEmpty)
	{
		if(strLen(%client.ticker))
		{
			%client.tickerEmpty = false;
			for(%i = 0; strLen(%mass) + %i < $tickerLength; %i++)
			{
				%client.ticker = " " @ %client.ticker;
			}
		}
		%client.bottomPrint($tickerFormatting @ %mass, 1, false);
	}
	else
	{
		%ticker = getSubStr(%client.ticker, 0, mClamp($tickerLength - strLen(%mass), 0, strLen(%client.ticker)));
		%client.ticker = getSubStr(%client.ticker, 1, strLen(%client.ticker));
		if(!strLen(%client.ticker))
		{
			%client.tickerEmpty = true;
		}
		%client.bottomPrint($tickerFormatting @ %mass @ %ticker, 1, false);
	}
	%client.agarHudSched = %client.schedule($tickerSpeed, agarHud);
}

// wip
function getLeaderboard()
{
	//compile a list of masses
	for(%i = 0; %i < clientGroup.getCount(); %i++)
	{
		%client = clientGroup.getObject(%i);
		%mass = 0;
		for(%j = 0; %j < %client.agarGroup.getCount(); %j++)
		{
			%mass+= %client.agarGroup.getObject(%j).mass;
		}
		%massRecord = setField(%massRecord, %i, mFloatLength(%mass, 2) SPC %client.name);
	}
	//pick out top 5 masses
	for(%j = 1; %j <= 5 && %j <= clientGroup.getCount(); %j++)
	{
		%topmass = 0;
		%top = "";
		for(%i = 0; %i < getFieldCount(%massRecord); %i++)
		{
			%field = getField(%massRecord, %i);
			if(firstWord(%field) > %topmass)
			{
				%topi = %i;
				%topmass = firstWord(%field);
				%top = %field;
			}
		}
		%massRecord = removeField(%massRecord, %topi);
		%leader[%j] = %top;
	}
	//assemble string
	for(%i = 1; %i < %j; %i++)
	{
		%str = %str @ " " @ %i @ ". " @ restWords(%leader[%i]) @ " (" @ firstWord(%leader[%i]) @ ") ";
	}
	return %str;
}

function periodicLeaderboard()
{
	%str = getLeaderboard();
	for(%i = 0; %i < clientGroup.getCount(); %i++)
	{
		%client = clientGroup.getObject(%i);
		%client.ticker = %client.ticker @ %str;
	}
	schedule($Agario::LeaderDisplayInterval, 0, periodicLeaderboard);
}

// #6.0

// #6.1
// spawn some food
for(%i = 0; %i < 250; %i++)
{
	schedule(500 + %i, 0, agarSpawnFood);
}